#ifndef UNITY_H_LOADED
#define UNITY_H_LOADED 1

/* From known-syntaxes.h ... */
/* KNOWN_SYNTAXES+ */
/**
 * Indicates the 'FITS' syntax.
 * Do not rely on the numerical value of this constant,
 * which may change between library versions without notice.
 */
#define UNITY_SYNTAX_FITS 0
/**
 * Indicates the 'OGIP' syntax.
 * Do not rely on the numerical value of this constant,
 * which may change between library versions without notice.
 */
#define UNITY_SYNTAX_OGIP 1
/**
 * Indicates the 'CDS' syntax.
 * Do not rely on the numerical value of this constant,
 * which may change between library versions without notice.
 */
#define UNITY_SYNTAX_CDS 2
#ifdef UNITY_INTERNAL
/* Number of syntaxes */
#define U_NSYNTAXES 3
#endif
/* KNOWN_SYNTAXES- */
/* ...end known-syntaxes.h */

/* From unitylib.h ... */
#ifndef UNITYLIB_H_LOADED
#define UNITYLIB_H_LOADED 1

/**
 * @file unity.h
 */

/* forward definition -- see unit-definition.h */
struct unit_definition;

/** @struct unit_struct
 * A component of a unit expression.  Precisely one of base_unit_def
 * or base_unit_string will be non-NULL.
 */
struct unit_struct {
    /**
     * The base-10 logarithm of the prefix in front of the unit.  For
     * example, this will be '-3' for the 'm' (milli-) prefix.
     */
    int prefix_power;           /* -24 to +24 */
    /**
     * The base unit of this unit specification, as a known unit.  If
     * this is not a known unit (in the syntax which parsed the
     * associated expression) then this will be NULL.
     */
    const struct unit_definition* base_unit_def;
    /**
     * The base unit of this unit specification.  This might be 'm'
     * for metres, 'Hz' for Hertz, and so on.
     */
    char* base_unit_string;     /* eg Hz or aangstrom */
    /**
     * The power to which the unit is raised.  The specification
     * 'mm^(-2)' will result in a power of '-2'.
     */
    float power;
    /** A pointer to the next unit in the sequence */
    struct unit_struct *next;
};
/** A clearer name for struct unit_struct */
typedef struct unit_struct Unit;

/** @struct UnitExpression
 * The parsed version of a unit string.
 */
typedef struct {
    /**
     * The base-10 log of the numerical factor at the head of the string. For some
     * parsers, this is restricted to round powers of ten.
     */
    float log_factor;
    /**
     * The sequence of units which comprise the parsed units specification.
     */
    Unit* unit_sequence;
} UnitExpression;

const UnitExpression* unity_parse_string(const char* unit_string, const int syntax);
/* unity_parse_error is defined in lex.lex */
/**
 * Retrieve the most recent lexing or parsing error.  Immediately after
 * the parser has returned with an error (that is function
 * 'unity_parse_string' has returned NULL) this function can be called
 * to obtain an explanation.
 * @sa unity_parse_string
 * @return a static string containing an explanation of a parse error
 */
const char* unity_parse_error(void);
const char** unity_parser_names(void);

int unity_identify_parser(const char* parser_name);
const char* unity_parser_name(const int parser_id);
void unity_free_expression(const UnitExpression* u);
const char* unity_version_string();
int unity_version_number();

/* Checking units */
#define UNITY_CHECK_RECOGNISED 1
#define UNITY_CHECK_RECOMMENDED 2
#define UNITY_CHECK_CONSTRAINTS 4
#define UNITY_CHECK_ALL UNITY_CHECK_RECOGNISED | UNITY_CHECK_RECOMMENDED | UNITY_CHECK_CONSTRAINTS
int unity_check_unit(const Unit* u, const int syntax, const int flags);
int unity_check_expression(const UnitExpression* ue, const int syntax, const int flags);

#ifdef UNITY_INTERNAL
/* Internal functions */
Unit* u_new_unit(const char* unitString, const float power, int syntax);
Unit* u_unit_append(Unit* u1, Unit* u2);
void u_receive_result(float, Unit* unit_sequence);
int u_prefix_to_power(const char pfx);
//char* u_copy_string(const char c, const char* s);
Unit* u_divide_units(Unit* num, Unit* den);
Unit* u_unit_reciprocal(Unit* u);
void u_free_unit(Unit* u);
#endif

#endif
/* ...end unitylib.h */

/* From writers.h ... */
#ifndef WRITERS_H_LOADED
#define WRITERS_H_LOADED 1

#include <stdio.h>

#ifdef UNITY_INTERNAL
typedef char* (*writef_ptr) (char*, int, const UnitExpression*);
#endif

char* unity_write_formatted(char* buf, int buflen, const UnitExpression*, const char*);
const char** unity_formatter_names(void);

#endif
/* ...end writers.h */

/* From unit-definitions.h ... */
#ifndef UNIT_DEFINITIONS_H
#define UNIT_DEFINITIONS_H 1


/* The functions exported by this module are the only ones which need
   access to the data structures dumped into this file, which are
   potentially subject to change.  The implementation of this header
   file is generated by ../grammar/ParseUnits.java, using the
   unit-definitions.c.template file in that directory.

   The structure elements with leading underscores are intended to be internal to
   this unit-definition.c module.
*/

struct unit_definition {
    const char* uri;            /* the name for this unit, as a URI */
    int _idx;                   /* the internal index for this unit, in the list of unit_definition entries */
    const char* name; /* the name of the unit, such as 'Metre' -- not the abbreviation 'm' */
    const char* type; /* type such as 'length', as a URI */
    const char* dimension; /* the dimensions of this unit, as a string, eg "M L-2" */
    const char* description; /* for example 'Hz = s-1', NULL if no notes present */
    const char* latex_form; /* an abbrev suitable for LaTeX, if different from a preferred syntax */
};
typedef struct unit_definition UnitDef;

#if UNITY_INTERNAL
/* A unit_representation indicates how a particular unit may be used in a particular syntax. */
struct unit_representation {
    const char* symbol;         /* a symbol for this unit, in one syntax */
    int _unit_index;            /* the index in the array of unit_definition entries */
    int si_prefixes_p;          /* true if this unit may be given SI prefixes */
    int is_deprecated_p;        /* true if this unit is deprecated */
    int is_preferred_p;         /* true if this unit is the preferred unit amongst some alternatives */
};
typedef struct unit_representation UnitRep;
UnitRep* u_get_unit_representation(const Unit*, const int syntax);
#endif

const UnitDef* unity_get_unit(const char* abbrev, int syntax);
const char* unity_get_syntax_name(int syntax);
const char* unity_get_unit_uri(UnitDef*);
const char* unity_get_unit_name(UnitDef*);
const char* unity_get_unit_type(UnitDef*);
const char* unity_get_unit_dimension(UnitDef*);
const char* unity_get_unit_description(UnitDef*);
const char* unity_get_unit_latex_form(UnitDef*);
#endif /* UNIT_DEFINITIONS_H */
/* ...end unit-definitions.h */
#endif
