/**
 * @file unity.h
 */

#include <stdio.h>              /* for NULL */
#include <stdlib.h>             /* for bsearch */
#include <string.h>             /* for strcmp */
#include <assert.h>

#define UNITY_INTERNAL 1
#include "unit-definitions.h"

static const char* known_syntaxes[] = { "fits", "ogip", "cds", };
/* Extract the following into unitylib.h */
/* KNOWN_SYNTAXES+ */
/**
 * Indicates the 'FITS' syntax.
 * Do not rely on the numerical value of this constant,
 * which may change between library versions without notice.
 */
#define UNITY_SYNTAX_FITS 0
/**
 * Indicates the 'OGIP' syntax.
 * Do not rely on the numerical value of this constant,
 * which may change between library versions without notice.
 */
#define UNITY_SYNTAX_OGIP 1
/**
 * Indicates the 'CDS' syntax.
 * Do not rely on the numerical value of this constant,
 * which may change between library versions without notice.
 */
#define UNITY_SYNTAX_CDS 2
#ifdef UNITY_INTERNAL
/* Number of syntaxes */
#define U_NSYNTAXES 3
#endif
/* KNOWN_SYNTAXES- */

/* List of all the known units */
static struct unit_definition unit_defs[] = {
  { "http://qudt.org/vocab/unit#Henry", 0,
    "Henry",
    "http://qudt.org/vocab/quantity#Inductance",
    "L ",
    NULL,
    NULL },
  { "http://bitbucket.org/nxg/unity/ns/unit#SolarMass", 1,
    "Solar mass",
    "http://qudt.org/vocab/quantity#Mass",
    "M ",
    "The mass of the sun is 1.9891e30 kg",
    "M\\solar" },
  { "http://qudt.org/vocab/unit#Parsec", 2,
    "Parsec",
    "http://qudt.org/vocab/quantity#Length",
    "L ",
    "The parsec (parallax of one arcsecond; symbol: pc) is a unit of length, equal to just under 31 trillion (31?1012) kilometres (about 19 trillion miles), 206265 AU, or about 3.26 light-years. The parsec measurement unit is used in astronomy. It is defined as the length of the adjacent side of an imaginary right triangle in space. The two dimensions that specify this triangle are the parallax angle (defined as 1 arcsecond) and the opposite side (defined as 1 astronomical unit (AU), the distance from the Earth to the Sun). Given these two measurements, along with the rules of trigonometry, the length of the adjacent side (the parsec) can be found. [Wikipedia]",
    NULL },
  { "http://qudt.org/vocab/unit#Siemens", 3,
    "Siemens",
    "http://qudt.org/vocab/quantity#ElectricConductivity",
    "L^-1 T ",
    NULL,
    NULL },
  { "http://bitbucket.org/nxg/unity/ns/unit#Pixel", 4,
    "pixel",
    "http://qudt.org/vocab/quantity#Dimensionless",
    "U ",
    "an element of a 2-D image",
    NULL },
  { "http://qudt.org/vocab/unit#AstronomicalUnit", 5,
    "Astronomical Unit",
    "http://qudt.org/vocab/quantity#Length",
    "L ",
    NULL,
    NULL },
  { "http://bitbucket.org/nxg/unity/ns/unit#SolarRadius", 6,
    "solRad",
    "http://qudt.org/vocab/quantity#Length",
    "L ",
    "6.9599e8 m",
    "R\\solar" },
  { "http://qudt.org/vocab/unit#ElectronVolt", 7,
    "Electron Volt",
    "http://qudt.org/vocab/quantity#EnergyAndWork",
    "L^2 M T^-2 ",
    NULL,
    NULL },
  { "http://qudt.org/vocab/unit#Farad", 8,
    "Farad",
    "http://qudt.org/vocab/quantity#Capacitance",
    "L^-1 T^2 ",
    NULL,
    NULL },
  { "http://qudt.org/vocab/unit#LightYear", 9,
    "Light Year",
    "http://qudt.org/vocab/quantity#Length",
    "L ",
    "A unit of length defining the distance, in meters, that light travels in a vacuum in one year.",
    NULL },
  { "http://qudt.org/vocab/unit#UnifiedAtomicMassUnit", 10,
    "Unified Atomic Mass Unit",
    "http://qudt.org/vocab/quantity#Mass",
    "M ",
    NULL,
    NULL },
  { "http://qudt.org/vocab/unit#Barn", 11,
    "Barn",
    "http://qudt.org/vocab/quantity#Area",
    "L^2 ",
    NULL,
    NULL },
  { "http://qudt.org/vocab/unit#Volt", 12,
    "Volt",
    "http://qudt.org/vocab/quantity#EnergyPerElectricCharge",
    "L^1.5 M^0.5 T^-2 ",
    NULL,
    NULL },
  { "http://bitbucket.org/nxg/unity/ns/unit#Crab", 13,
    "Crab",
    "http://bitbucket/org/nxg/unity/ns/quantity#SpecificIntensity",
    "M T^-2 ",
    "The flux density, relative to that of the Crab.\nMentioned in the OGIP standard, but strongly disrecommended.\nThe OGIP standard permits only 'm' as a prefix.",
    NULL },
  { "http://qudt.org/vocab/unit#MinuteTime", 14,
    "Minute Time",
    "http://qudt.org/vocab/quantity#Time",
    "T ",
    NULL,
    NULL },
  { "http://qudt.org/vocab/unit#Number", 15,
    "Number",
    "http://qudt.org/vocab/quantity#Dimensionless",
    "U ",
    NULL,
    NULL },
  { "http://qudt.org/vocab/unit#DegreeAngle", 16,
    "Degree Angle",
    "http://qudt.org/vocab/quantity#PlaneAngle",
    "U ",
    NULL,
    NULL },
  { "http://qudt.org/vocab/unit#ArcMinute", 17,
    "Arc Minute",
    "http://qudt.org/vocab/quantity#PlaneAngle",
    "U ",
    NULL,
    NULL },
  { "http://qudt.org/vocab/unit#Day", 18,
    "Day",
    "http://qudt.org/vocab/quantity#Time",
    "T ",
    "Mean solar day",
    NULL },
  { "http://bitbucket.org/nxg/unity/ns/unit#Jansky", 19,
    "Jy",
    "http://bitbucket/org/nxg/unity/ns/quantity#SpecificIntensity",
    "M T^-2 ",
    "10^-26 W m-2 Hz-1",
    NULL },
  { "http://bitbucket.org/nxg/unity/ns/unit#Beam", 20,
    "beam",
    "http://qudt.org/vocab/quantity#Dimensionless",
    "U ",
    "as in Jy/beam",
    NULL },
  { "http://qudt.org/vocab/unit#Joule", 21,
    "Joule",
    "http://qudt.org/vocab/quantity#EnergyAndWork",
    "L^2 M T^-2 ",
    NULL,
    NULL },
  { "http://bitbucket.org/nxg/unity/ns/unit#Voxel", 22,
    "voxel",
    "http://qudt.org/vocab/quantity#Dimensionless",
    "U ",
    "The 3-D analogue of a pixel",
    NULL },
  { "http://qudt.org/vocab/unit#Newton", 23,
    "Newton",
    "http://qudt.org/vocab/quantity#Force",
    "L M T^-2 ",
    NULL,
    NULL },
  { "http://qudt.org/vocab/unit#ArcSecond", 24,
    "Arc Second",
    "http://qudt.org/vocab/quantity#PlaneAngle",
    "U ",
    NULL,
    NULL },
  { "http://qudt.org/vocab/unit#Mole", 25,
    "Mole",
    "http://qudt.org/vocab/quantity#AmountOfSubstance",
    "N ",
    NULL,
    NULL },
  { "http://bitbucket.org/nxg/unity/ns/unit#Rydberg", 26,
    "Ry",
    "http://qudt.org/vocab/quantity#EnergyAndWork",
    "L^2 M T^-2 ",
    "Half of the Hartree energy, defined as 13.605692 eV.  See http://en.wikipedia.org/wiki/Hartree",
    NULL },
  { "http://bitbucket.org/nxg/unity/ns/unit#MilliArcSecond", 27,
    "mas",
    "http://qudt.org/vocab/quantity#PlaneAngle",
    "U ",
    "1/1000 of an arc-second, for use in those syntaxes which do not allow SI prefixes of arcsec",
    NULL },
  { "http://qudt.org/vocab/unit#Erg", 28,
    "Erg",
    "http://qudt.org/vocab/quantity#EnergyAndWork",
    "L^2 M T^-2 ",
    NULL,
    NULL },
  { "http://qudt.org/vocab/unit#Percent", 29,
    "Percent",
    "http://qudt.org/vocab/quantity#DimensionlessRatio",
    "U ",
    NULL,
    NULL },
  { "http://bitbucket.org/nxg/unity/ns/unit#StellarMagnitude", 30,
    "mag",
    "http://qudt.org/vocab/quantity#Dimensionless",
    "U ",
    "Log of a ratio of intensities.  See http://en.wikipedia.org/wiki/Magnitude_(astronomy)",
    NULL },
  { "http://bitbucket.org/nxg/unity/ns/unit#DetectorChannel", 31,
    "chan",
    "http://qudt.org/vocab/quantity#Dimensionless",
    "U ",
    "One channel in a detector of some type",
    NULL },
  { "http://bitbucket.org/nxg/unity/ns/unit#ADU", 32,
    "ADU",
    "http://qudt.org/vocab/quantity#Dimensionless",
    "U ",
    "Same as channel???",
    NULL },
  { "http://bitbucket.org/nxg/unity/ns/unit#Photon", 33,
    "ph",
    "http://qudt.org/vocab/quantity#Dimensionless",
    "U ",
    "as in ev/photon",
    NULL },
  { "http://qudt.org/vocab/unit#Radian", 34,
    "Radian",
    "http://qudt.org/vocab/quantity#PlaneAngle",
    "U ",
    "The radian is the standard unit of angular measure, used in many areas of mathematics. It describes the plane angle subtended by a circular arc as the length of the arc divided by the radius of the arc. The unit was formerly a SI supplementary unit, but this category was abolished in 1995 and the radian is now considered a SI derived unit. The SI unit of solid angle measurement is the steradian.\nThe radian is represented by the symbol \"rad\" or, more rarely, by the superscript c (for \"circular measure\"). For example, an angle of 1.2 radians would be written as \"1.2 rad\" or \"1.2c\" (the second symbol is often mistaken for a degree: \"1.2\"). As the ratio of two lengths, the radian is a \"pure number\" that needs no unit symbol, and in mathematical writing the symbol \"rad\" is almost always omitted. In the absence of any symbol radians are assumed, and when degrees are meant the symbol  is used. [Wikipedia]",
    NULL },
  { "http://bitbucket.org/nxg/unity/ns/unit#JulianYear", 35,
    "Julian year",
    "http://qudt.org/vocab/quantity#Time",
    "T ",
    "31 557 600s (365.25d), peta a (Pa) forbidden",
    NULL },
  { "http://qudt.org/vocab/unit#Hertz", 36,
    "Hertz",
    "http://qudt.org/vocab/quantity#Frequency",
    "T^-1 ",
    NULL,
    NULL },
  { "http://qudt.org/vocab/unit#Watt", 37,
    "Watt",
    "http://qudt.org/vocab/quantity#HeatFlowRate",
    "L^2 M T^-3 ",
    NULL,
    NULL },
  { "http://bitbucket.org/nxg/unity/ns/unit#DistributionBin", 38,
    "bin",
    "http://qudt.org/vocab/quantity#Dimensionless",
    "U ",
    "One division within a distribution",
    NULL },
  { "http://qudt.org/vocab/unit#Ohm", 39,
    "Ohm",
    "http://qudt.org/vocab/quantity#Resistance",
    "L T^-1 ",
    NULL,
    "\\ohm" },
  { "http://qudt.org/vocab/unit#Candela", 40,
    "Candela",
    "http://qudt.org/vocab/quantity#LuminousIntensity",
    "J ",
    NULL,
    NULL },
  { "http://qudt.org/vocab/unit#Angstrom", 41,
    "Angstrom",
    "http://qudt.org/vocab/quantity#Length",
    "L ",
    "The ngstrm or angstrom (symbol ) (the latter spelling, without diacritics, is now usually used in English) (pronounced /??str?m/; Swedish: [???strm]) is an internationally recognized unit of length equal to 0.1 nanometre or 1?10?10 metres. It is named after Anders Jonas ngstrm. Although accepted for use, it is not formally defined within the International System of Units(SI). (That article lists the units that are so defined.) The ngstrm is often used in the natural sciences to express the sizes of atoms, lengths of chemical bonds and the wavelengths of electromagnetic radiation, and in technology for the dimensions of parts of integrated circuits. It is also commonly used in structural biology. [Wikipedia]",
    "\\angstrom" },
  { "http://qudt.org/vocab/unit#Gauss", 42,
    "Gauss",
    "http://qudt.org/vocab/quantity#MagneticField",
    "L^-0.5 M^0.5 T^-1 ",
    NULL,
    NULL },
  { "http://qudt.org/vocab/unit#Meter", 43,
    "Metre",
    "http://qudt.org/vocab/quantity#Length",
    "L ",
    NULL,
    NULL },
  { "http://qudt.org/vocab/unit#Lumen", 44,
    "Lumen",
    "http://qudt.org/vocab/quantity#LuminousFlux",
    "U J ",
    NULL,
    NULL },
  { "http://qudt.org/vocab/unit#Bit", 45,
    "Bit",
    "http://qudt.org/vocab/quantity#InformationEntropy",
    "U ",
    "In information theory, a bit is the amount of information that, on average, can be stored in a discrete bit. It is thus the amount of information carried by a choice between two equally likely outcomes. One bit corresponds to about 0.693 nats (ln(2)), or 0.301 hartleys (log10(2)).",
    NULL },
  { "http://qudt.org/vocab/unit#Kelvin", 46,
    "Kelvin",
    "http://qudt.org/vocab/quantity#ThermodynamicTemperature",
    "? ",
    NULL,
    NULL },
  { "http://qudt.org/vocab/unit#Weber", 47,
    "Weber",
    "http://qudt.org/vocab/quantity#MagneticFlux",
    "L^1.5 M^0.5 T^-1 ",
    NULL,
    NULL },
  { "http://qudt.org/vocab/unit#Ampere", 48,
    "Ampere",
    "http://qudt.org/vocab/quantity#ElectricCurrent",
    "L^0.5 M^0.5 T^-1 ",
    NULL,
    NULL },
  { "http://qudt.org/vocab/unit#Hour", 49,
    "Hour",
    "http://qudt.org/vocab/quantity#Time",
    "T ",
    NULL,
    NULL },
  { "http://qudt.org/vocab/unit#Tesla", 50,
    "Tesla",
    "http://qudt.org/vocab/quantity#MagneticField",
    "L^-0.5 M^0.5 T^-1 ",
    NULL,
    NULL },
  { "http://qudt.org/vocab/unit#Lux", 51,
    "Lux",
    "http://qudt.org/vocab/quantity#LuminousFluxPerArea",
    "U L^-2 J ",
    NULL,
    NULL },
  { "http://qudt.org/vocab/unit#Steradian", 52,
    "Steradian",
    "http://qudt.org/vocab/quantity#SolidAngle",
    "U ",
    "The steradian (symbol: sr) is the SI unit of solid angle. It is used to describe two-dimensional angular spans in three-dimensional space, analogous to the way in which the radian describes angles in a plane.",
    NULL },
  { "http://qudt.org/vocab/unit#Debye", 53,
    "Debye",
    "http://qudt.org/vocab/quantity#ElectricDipoleMoment",
    "L T I ",
    NULL,
    NULL },
  { "http://qudt.org/vocab/unit#SecondTime", 54,
    "Second",
    "http://qudt.org/vocab/quantity#Time",
    "T ",
    NULL,
    NULL },
  { "http://qudt.org/vocab/unit#Gram", 55,
    "Gramme",
    "http://qudt.org/vocab/quantity#Mass",
    "M ",
    NULL,
    NULL },
  { "http://qudt.org/vocab/unit#Coulomb", 56,
    "Coulomb",
    "http://qudt.org/vocab/quantity#ElectricCharge",
    "L^0.5 M^0.5 ",
    NULL,
    NULL },
  { "http://bitbucket.org/nxg/unity/ns/unit#Rayleigh", 57,
    "R",
    "http://bitbucket.org/nxg/unity/ns/unit#PhotonFlux",
    "L^-2 T-2",
    "Unit of photon flux density: Wikipedia has units of 10^10 photons/m2/column/s.\nThe FITS spec has units of 1e10/(4PI) photons m-2 s-2 sr-1.\nSee http://en.wikipedia.org/wiki/Rayleigh_(unit).\nNo, I don't know how to express its dimensions in the QUDT system!\nWe go with the FITS definition here.",
    NULL },
  { "http://qudt.org/vocab/unit#Byte", 58,
    "Byte",
    "http://qudt.org/vocab/quantity#InformationEntropy",
    "U ",
    NULL,
    NULL },
  { "http://bitbucket.org/nxg/unity/ns/unit#SolarLuminosity", 59,
    "solLum",
    "http://qudt.org/vocab/quantity#Power",
    "L^2 M T^-3 ",
    "3.8268e26 W",
    "L\\solar" },
  { "http://qudt.org/vocab/unit#Pascal", 60,
    "Pascal",
    "http://qudt.org/vocab/quantity#ForcePerArea",
    "L^-1 M T^-2 ",
    NULL,
    NULL },
};
static int unit_defs_length = 61;


// Reverse lookups, per-syntax maps going from symbols to UnitDefinitions

/* A lookup table, mapping symbols to an index in the unit_representation array above.
   There are three arrays in here, separatedly ordered by symbol,
   with the start offsets indicated by the unit_rep_per_syntax array below. */
static struct unit_representation unit_reps[] = {

  // unit representations for syntax fits
  { "A", 48, 1, 0, 0 }, // http://qudt.org/vocab/unit#Ampere
  { "AU", 5, 0, 0, 0 }, // http://qudt.org/vocab/unit#AstronomicalUnit
  { "Angstrom", 41, 0, 1, 0 }, // http://qudt.org/vocab/unit#Angstrom
  { "C", 56, 1, 0, 0 }, // http://qudt.org/vocab/unit#Coulomb
  { "D", 53, 0, 0, 0 }, // http://qudt.org/vocab/unit#Debye
  { "F", 8, 1, 0, 0 }, // http://qudt.org/vocab/unit#Farad
  { "G", 42, 1, 1, 0 }, // http://qudt.org/vocab/unit#Gauss
  { "H", 0, 1, 0, 0 }, // http://qudt.org/vocab/unit#Henry
  { "Hz", 36, 1, 0, 0 }, // http://qudt.org/vocab/unit#Hertz
  { "J", 21, 1, 0, 0 }, // http://qudt.org/vocab/unit#Joule
  { "Jy", 19, 1, 0, 0 }, // http://bitbucket.org/nxg/unity/ns/unit#Jansky
  { "K", 46, 1, 0, 0 }, // http://qudt.org/vocab/unit#Kelvin
  { "N", 23, 1, 0, 0 }, // http://qudt.org/vocab/unit#Newton
  { "Ohm", 39, 1, 0, 0 }, // http://qudt.org/vocab/unit#Ohm
  { "Pa", 60, 1, 0, 0 }, // http://qudt.org/vocab/unit#Pascal
  { "R", 57, 1, 0, 0 }, // http://bitbucket.org/nxg/unity/ns/unit#Rayleigh
  { "Ry", 26, 0, 0, 0 }, // http://bitbucket.org/nxg/unity/ns/unit#Rydberg
  { "S", 3, 1, 0, 0 }, // http://qudt.org/vocab/unit#Siemens
  { "T", 50, 1, 0, 0 }, // http://qudt.org/vocab/unit#Tesla
  { "V", 12, 1, 0, 0 }, // http://qudt.org/vocab/unit#Volt
  { "W", 37, 1, 0, 0 }, // http://qudt.org/vocab/unit#Watt
  { "Wb", 47, 1, 0, 0 }, // http://qudt.org/vocab/unit#Weber
  { "a", 35, 1, 0, 0 }, // http://bitbucket.org/nxg/unity/ns/unit#JulianYear
  { "adu", 32, 0, 0, 0 }, // http://bitbucket.org/nxg/unity/ns/unit#ADU
  { "arcmin", 17, 0, 0, 0 }, // http://qudt.org/vocab/unit#ArcMinute
  { "arcsec", 24, 0, 0, 0 }, // http://qudt.org/vocab/unit#ArcSecond
  { "barn", 11, 1, 1, 0 }, // http://qudt.org/vocab/unit#Barn
  { "beam", 20, 0, 0, 0 }, // http://bitbucket.org/nxg/unity/ns/unit#Beam
  { "bin", 38, 0, 0, 0 }, // http://bitbucket.org/nxg/unity/ns/unit#DistributionBin
  { "bit", 45, 1, 0, 0 }, // http://qudt.org/vocab/unit#Bit
  { "byte", 58, 1, 0, 0 }, // http://qudt.org/vocab/unit#Byte
  { "cd", 40, 1, 0, 0 }, // http://qudt.org/vocab/unit#Candela
  { "chan", 31, 0, 0, 0 }, // http://bitbucket.org/nxg/unity/ns/unit#DetectorChannel
  { "count", 15, 0, 0, 0 }, // http://qudt.org/vocab/unit#Number
  { "ct", 15, 0, 0, 0 }, // http://qudt.org/vocab/unit#Number
  { "d", 18, 0, 0, 0 }, // http://qudt.org/vocab/unit#Day
  { "deg", 16, 0, 0, 0 }, // http://qudt.org/vocab/unit#DegreeAngle
  { "eV", 7, 1, 0, 0 }, // http://qudt.org/vocab/unit#ElectronVolt
  { "erg", 28, 0, 1, 0 }, // http://qudt.org/vocab/unit#Erg
  { "g", 55, 1, 0, 0 }, // http://qudt.org/vocab/unit#Gram
  { "h", 49, 0, 0, 0 }, // http://qudt.org/vocab/unit#Hour
  { "lm", 44, 1, 0, 0 }, // http://qudt.org/vocab/unit#Lumen
  { "lx", 51, 1, 0, 0 }, // http://qudt.org/vocab/unit#Lux
  { "lyr", 9, 0, 0, 0 }, // http://qudt.org/vocab/unit#LightYear
  { "m", 43, 1, 0, 0 }, // http://qudt.org/vocab/unit#Meter
  { "mag", 30, 1, 0, 0 }, // http://bitbucket.org/nxg/unity/ns/unit#StellarMagnitude
  { "mas", 27, 0, 0, 0 }, // http://bitbucket.org/nxg/unity/ns/unit#MilliArcSecond
  { "min", 14, 0, 0, 0 }, // http://qudt.org/vocab/unit#MinuteTime
  { "mol", 25, 1, 0, 0 }, // http://qudt.org/vocab/unit#Mole
  { "pc", 2, 1, 0, 0 }, // http://qudt.org/vocab/unit#Parsec
  { "ph", 33, 0, 0, 0 }, // http://bitbucket.org/nxg/unity/ns/unit#Photon
  { "photon", 33, 0, 0, 1 }, // http://bitbucket.org/nxg/unity/ns/unit#Photon
  { "pix", 4, 0, 0, 0 }, // http://bitbucket.org/nxg/unity/ns/unit#Pixel
  { "pixel", 4, 0, 0, 1 }, // http://bitbucket.org/nxg/unity/ns/unit#Pixel
  { "rad", 34, 1, 0, 0 }, // http://qudt.org/vocab/unit#Radian
  { "s", 54, 1, 0, 0 }, // http://qudt.org/vocab/unit#SecondTime
  { "solLum", 59, 0, 0, 0 }, // http://bitbucket.org/nxg/unity/ns/unit#SolarLuminosity
  { "solMass", 1, 0, 0, 0 }, // http://bitbucket.org/nxg/unity/ns/unit#SolarMass
  { "solRad", 6, 0, 0, 0 }, // http://bitbucket.org/nxg/unity/ns/unit#SolarRadius
  { "sr", 52, 1, 0, 0 }, // http://qudt.org/vocab/unit#Steradian
  { "u", 10, 0, 0, 0 }, // http://qudt.org/vocab/unit#UnifiedAtomicMassUnit
  { "voxel", 22, 0, 0, 0 }, // http://bitbucket.org/nxg/unity/ns/unit#Voxel
  { "yr", 35, 1, 0, 1 }, // http://bitbucket.org/nxg/unity/ns/unit#JulianYear

  // unit representations for syntax ogip
  { "A", 48, 1, 0, 0 }, // http://qudt.org/vocab/unit#Ampere
  { "AU", 5, 0, 0, 0 }, // http://qudt.org/vocab/unit#AstronomicalUnit
  { "Angstrom", 41, 0, 0, 0 }, // http://qudt.org/vocab/unit#Angstrom
  { "C", 56, 1, 0, 0 }, // http://qudt.org/vocab/unit#Coulomb
  { "Crab", 13, 1, 0, 0 }, // http://bitbucket.org/nxg/unity/ns/unit#Crab
  { "F", 8, 1, 0, 0 }, // http://qudt.org/vocab/unit#Farad
  { "G", 42, 0, 0, 0 }, // http://qudt.org/vocab/unit#Gauss
  { "H", 0, 1, 0, 0 }, // http://qudt.org/vocab/unit#Henry
  { "Hz", 36, 1, 0, 0 }, // http://qudt.org/vocab/unit#Hertz
  { "J", 21, 1, 0, 0 }, // http://qudt.org/vocab/unit#Joule
  { "Jy", 19, 1, 0, 0 }, // http://bitbucket.org/nxg/unity/ns/unit#Jansky
  { "K", 46, 1, 0, 0 }, // http://qudt.org/vocab/unit#Kelvin
  { "N", 23, 1, 0, 0 }, // http://qudt.org/vocab/unit#Newton
  { "Pa", 60, 1, 0, 0 }, // http://qudt.org/vocab/unit#Pascal
  { "S", 3, 1, 0, 0 }, // http://qudt.org/vocab/unit#Siemens
  { "T", 50, 1, 0, 0 }, // http://qudt.org/vocab/unit#Tesla
  { "V", 12, 1, 0, 0 }, // http://qudt.org/vocab/unit#Volt
  { "W", 37, 1, 0, 0 }, // http://qudt.org/vocab/unit#Watt
  { "Wb", 47, 1, 0, 0 }, // http://qudt.org/vocab/unit#Weber
  { "arcmin", 17, 0, 0, 0 }, // http://qudt.org/vocab/unit#ArcMinute
  { "arcsec", 24, 0, 0, 0 }, // http://qudt.org/vocab/unit#ArcSecond
  { "barn", 11, 0, 0, 0 }, // http://qudt.org/vocab/unit#Barn
  { "bin", 38, 0, 0, 0 }, // http://bitbucket.org/nxg/unity/ns/unit#DistributionBin
  { "byte", 58, 0, 0, 0 }, // http://qudt.org/vocab/unit#Byte
  { "cd", 40, 1, 0, 0 }, // http://qudt.org/vocab/unit#Candela
  { "chan", 31, 0, 0, 0 }, // http://bitbucket.org/nxg/unity/ns/unit#DetectorChannel
  { "count", 15, 0, 0, 0 }, // http://qudt.org/vocab/unit#Number
  { "d", 18, 0, 0, 0 }, // http://qudt.org/vocab/unit#Day
  { "deg", 16, 0, 0, 0 }, // http://qudt.org/vocab/unit#DegreeAngle
  { "eV", 7, 1, 0, 0 }, // http://qudt.org/vocab/unit#ElectronVolt
  { "erg", 28, 0, 0, 0 }, // http://qudt.org/vocab/unit#Erg
  { "g", 55, 1, 0, 0 }, // http://qudt.org/vocab/unit#Gram
  { "h", 49, 0, 0, 0 }, // http://qudt.org/vocab/unit#Hour
  { "lm", 44, 1, 0, 0 }, // http://qudt.org/vocab/unit#Lumen
  { "lx", 51, 1, 0, 0 }, // http://qudt.org/vocab/unit#Lux
  { "lyr", 9, 0, 0, 0 }, // http://qudt.org/vocab/unit#LightYear
  { "m", 43, 1, 0, 0 }, // http://qudt.org/vocab/unit#Meter
  { "mag", 30, 0, 0, 0 }, // http://bitbucket.org/nxg/unity/ns/unit#StellarMagnitude
  { "min", 14, 0, 0, 0 }, // http://qudt.org/vocab/unit#MinuteTime
  { "mol", 25, 1, 0, 0 }, // http://qudt.org/vocab/unit#Mole
  { "ohm", 39, 1, 0, 0 }, // http://qudt.org/vocab/unit#Ohm
  { "pc", 2, 1, 0, 0 }, // http://qudt.org/vocab/unit#Parsec
  { "photon", 33, 0, 0, 0 }, // http://bitbucket.org/nxg/unity/ns/unit#Photon
  { "pixel", 4, 0, 0, 0 }, // http://bitbucket.org/nxg/unity/ns/unit#Pixel
  { "rad", 34, 1, 0, 0 }, // http://qudt.org/vocab/unit#Radian
  { "s", 54, 1, 0, 0 }, // http://qudt.org/vocab/unit#SecondTime
  { "sr", 52, 1, 0, 0 }, // http://qudt.org/vocab/unit#Steradian
  { "voxel", 22, 0, 0, 0 }, // http://bitbucket.org/nxg/unity/ns/unit#Voxel
  { "yr", 35, 0, 0, 0 }, // http://bitbucket.org/nxg/unity/ns/unit#JulianYear

  // unit representations for syntax cds
  { "%", 29, 0, 0, 0 }, // http://qudt.org/vocab/unit#Percent
  { "A", 48, 1, 0, 0 }, // http://qudt.org/vocab/unit#Ampere
  { "AU", 5, 0, 0, 0 }, // http://qudt.org/vocab/unit#AstronomicalUnit
  { "C", 56, 1, 0, 0 }, // http://qudt.org/vocab/unit#Coulomb
  { "D", 53, 0, 0, 0 }, // http://qudt.org/vocab/unit#Debye
  { "F", 8, 1, 0, 0 }, // http://qudt.org/vocab/unit#Farad
  { "H", 0, 1, 0, 0 }, // http://qudt.org/vocab/unit#Henry
  { "Hz", 36, 1, 0, 0 }, // http://qudt.org/vocab/unit#Hertz
  { "J", 21, 1, 0, 0 }, // http://qudt.org/vocab/unit#Joule
  { "Jy", 19, 1, 0, 0 }, // http://bitbucket.org/nxg/unity/ns/unit#Jansky
  { "K", 46, 1, 0, 0 }, // http://qudt.org/vocab/unit#Kelvin
  { "N", 23, 1, 0, 0 }, // http://qudt.org/vocab/unit#Newton
  { "Ohm", 39, 1, 0, 0 }, // http://qudt.org/vocab/unit#Ohm
  { "Pa", 60, 1, 0, 0 }, // http://qudt.org/vocab/unit#Pascal
  { "Ry", 26, 1, 0, 0 }, // http://bitbucket.org/nxg/unity/ns/unit#Rydberg
  { "S", 3, 1, 0, 0 }, // http://qudt.org/vocab/unit#Siemens
  { "T", 50, 1, 0, 0 }, // http://qudt.org/vocab/unit#Tesla
  { "V", 12, 1, 0, 0 }, // http://qudt.org/vocab/unit#Volt
  { "W", 37, 1, 0, 0 }, // http://qudt.org/vocab/unit#Watt
  { "Wb", 47, 1, 0, 0 }, // http://qudt.org/vocab/unit#Weber
  { "a", 35, 1, 0, 0 }, // http://bitbucket.org/nxg/unity/ns/unit#JulianYear
  { "arcmin", 17, 0, 0, 0 }, // http://qudt.org/vocab/unit#ArcMinute
  { "arcsec", 24, 1, 0, 0 }, // http://qudt.org/vocab/unit#ArcSecond
  { "barn", 11, 1, 0, 0 }, // http://qudt.org/vocab/unit#Barn
  { "bit", 45, 1, 0, 0 }, // http://qudt.org/vocab/unit#Bit
  { "byte", 58, 1, 0, 0 }, // http://qudt.org/vocab/unit#Byte
  { "cd", 40, 1, 0, 0 }, // http://qudt.org/vocab/unit#Candela
  { "ct", 15, 0, 0, 0 }, // http://qudt.org/vocab/unit#Number
  { "d", 18, 0, 0, 0 }, // http://qudt.org/vocab/unit#Day
  { "deg", 16, 0, 0, 0 }, // http://qudt.org/vocab/unit#DegreeAngle
  { "eV", 7, 1, 0, 0 }, // http://qudt.org/vocab/unit#ElectronVolt
  { "g", 55, 1, 0, 0 }, // http://qudt.org/vocab/unit#Gram
  { "h", 49, 0, 0, 0 }, // http://qudt.org/vocab/unit#Hour
  { "lm", 44, 1, 0, 0 }, // http://qudt.org/vocab/unit#Lumen
  { "lx", 51, 1, 0, 0 }, // http://qudt.org/vocab/unit#Lux
  { "m", 43, 1, 0, 0 }, // http://qudt.org/vocab/unit#Meter
  { "mag", 30, 1, 0, 0 }, // http://bitbucket.org/nxg/unity/ns/unit#StellarMagnitude
  { "mas", 27, 0, 0, 0 }, // http://bitbucket.org/nxg/unity/ns/unit#MilliArcSecond
  { "min", 14, 0, 0, 0 }, // http://qudt.org/vocab/unit#MinuteTime
  { "mol", 25, 1, 0, 0 }, // http://qudt.org/vocab/unit#Mole
  { "pc", 2, 1, 0, 0 }, // http://qudt.org/vocab/unit#Parsec
  { "pix", 4, 0, 0, 0 }, // http://bitbucket.org/nxg/unity/ns/unit#Pixel
  { "rad", 34, 1, 0, 0 }, // http://qudt.org/vocab/unit#Radian
  { "s", 54, 1, 0, 0 }, // http://qudt.org/vocab/unit#SecondTime
  { "solLum", 59, 0, 0, 0 }, // http://bitbucket.org/nxg/unity/ns/unit#SolarLuminosity
  { "solMass", 1, 0, 0, 0 }, // http://bitbucket.org/nxg/unity/ns/unit#SolarMass
  { "solRad", 6, 0, 0, 0 }, // http://bitbucket.org/nxg/unity/ns/unit#SolarRadius
  { "sr", 52, 1, 0, 0 }, // http://qudt.org/vocab/unit#Steradian
  { "yr", 35, 1, 0, 1 }, // http://bitbucket.org/nxg/unity/ns/unit#JulianYear
};
struct { int map_start; int map_length; } unit_rep_per_syntax[] = {
  { 0, 63 },
  { 63, 49 },
  { 112, 49 },
};

/* A mapping, for each syntax, from unit number (index in the unit_defs array above)
   to representation (index in the unit_reps array above) */
static int unit_to_representation[3][61] = {
  { 7, 57, 49, 17, 53, 1, 58, 37, 5, 43, 60, 26, 19, -1, 47, 34, 36, 24, 35, 10, 27, 9, 61, 12, 25, 48, 16, 46, 38, -1, 45, 32, 23, 51, 54, 62, 8, 20, 28, 13, 31, 2, 6, 44, 41, 29, 11, 21, 0, 40, 18, 42, 59, 4, 55, 39, 3, 15, 30, 56, 14, },
  { 70, -1, 104, 77, 106, 64, -1, 92, 68, 98, -1, 84, 79, 67, 101, 89, 91, 82, 90, 73, -1, 72, 110, 75, 83, 102, -1, -1, 93, -1, 100, 88, -1, 105, 107, 111, 71, 80, 85, 103, 87, 65, 69, 99, 96, -1, 74, 81, 63, 95, 78, 97, 109, -1, 108, 94, 66, -1, 86, -1, 76, },
  { 118, 157, 152, 127, 153, 114, 158, 142, 117, -1, -1, 135, 129, -1, 150, 139, 141, 133, 140, 121, -1, 120, -1, 123, 134, 151, 126, 149, -1, 112, 148, -1, -1, -1, 154, 160, 119, 130, -1, 124, 138, -1, -1, 147, 145, 136, 122, 131, 113, 144, 128, 146, 159, 116, 155, 143, 115, -1, 137, 156, 125, },
};

static int unit_rep_compare(const void* key, const void* test)
{
    return strcmp((char*)key, ((UnitRep*)test)->symbol);
}

/**
 * Retrieves a unit definition.  Returns NULL if the abbreviation is
 * not recognised in the indicated syntax (or if the syntax is invalid)
 * @param abbrev a unit abbreviation, for example "m" for "metre"
 * @param syntax one of the UNITY_SYNTAX_* constants
 * @return a pointer to a the unit definition which this abbreviation corresponds to
 */
const UnitDef* unity_get_unit(const char* abbrev, const int syntax) 
{
    UnitRep *map;
    UnitRep *one_unit_rep;
    int maplen;
    int rep_index;
    
    if (syntax < 0 || syntax >= U_NSYNTAXES) {
        return NULL;
    }
    map = &unit_reps[unit_rep_per_syntax[syntax].map_start];
    maplen = unit_rep_per_syntax[syntax].map_length;
    
    one_unit_rep = (UnitRep*)bsearch((void*)abbrev,
                                     (void*)map, maplen, sizeof(UnitRep),
                                     &unit_rep_compare);
    if (one_unit_rep != NULL) {
        rep_index = one_unit_rep->_unit_index;
        return &unit_defs[rep_index];
    } else {
        return NULL;
    }
}
UnitRep* u_get_unit_representation(const Unit* u, const int syntax)
{
    const UnitDef* d = u->base_unit_def;
    int unit_rep_index;

    if (syntax < 0 || syntax >= U_NSYNTAXES) {
        fprintf(stderr, "unity_check_unit: Illegal syntax specifier\n");
        return NULL;            /* JUMP OUT */
    }
    if (d == NULL) {
        // this unit doesn't have a base_unit_def -- it wasn't recognised at parse time
        return NULL;            /* JUMP OUT */
    }
    assert(d->_idx >= 0 && d->_idx <unit_defs_length);

    unit_rep_index = unit_to_representation[syntax][d->_idx];
    if (unit_rep_index < 0) {
        return NULL;
    } else {
        return &unit_reps[unit_rep_index];
    }
}

/**
 * Looks up the name of the syntax with the given index.
 * @param syntax one of the constants UNITY_SYNTAX_FITS, ...
 * @returns a string name for the syntax, or NULL if this is an unrecognised syntax
 */
const char* unity_get_syntax_name(const int syntax) 
{
    if (syntax >= 0 && syntax < U_NSYNTAXES) {
        return known_syntaxes[syntax];
    } else {
        return NULL;
    }
}

/**
 * Returns the name of this unit, as a URI.  These unit definitions
 * follow the <a href='http://qudt.org'>QUDT</a> framework, though
 * they are not restricted to the units described there.  
 *
 * <p>This framework also supports the definition of unit kinds and
 * dimensions.
 * @returns a URI as a string
 */
const char* unity_get_unit_uri(UnitDef* d)
{
    return (d == NULL ? NULL : d->uri);
}
/**
 * Returns the name of the unit.  This is a human-readable name such
 * as 'Metre' or 'Julian year', not the abbreviation 'm'
 */
const char* unity_get_unit_name(UnitDef* d)
{
    return (d == NULL ? NULL : d->name);
}
/**
 * Returns the type of the unit, as a URI naming the 'kind' of thing
 * this measures.  This indicates 'Length' or 'Capacitance'.
 */
const char* unity_get_unit_type(UnitDef* d)
{
    return (d == NULL ? NULL : d->type);
}
/**
 * Returns the dimensions of the unit
 *
 * <p>A dimensions string consists of a sequence of capital letter
 * dimensions, and powers, for example "M L2T-2" for the
 * dimensions of "kg m2 s-2".
 *
 * <p>The known dimensions are
 * <table>
 * <tr><th>Symbol<th>Name<th>SI base unit
 * <tr><td>U<td>Dimensionless<td>Unity
 * <tr><td>L<td>Length<td>Metre
 * <tr><td>M<td>Mass<td>Kilogramme
 * <tr><td>T<td>Time<td>Second
 * <tr><td>I<td>Electric current<td>Ampère
 * <tr><td>Θ<td>Thermodynamic temperature<td>Kelvin
 * <tr><td>N<td>Amount of substance<td>Mole
 * <tr><td>J<td>Luminous intensity<td>Candela
 * </table>
 */
const char* unity_get_unit_dimension(UnitDef* d)
{
    return (d == NULL ? NULL : d->dimension);
}
/**
 * Returns a description of the unit.  This may be NULL if there is
 * nothing more to be said beyond the unit name.
 */
const char* unity_get_unit_description(UnitDef* d)
{
    return (d == NULL ? NULL : d->description);
}
/**
 * A LaTeX version of the unit symbol, or there is one defined.
 */
const char* unity_get_unit_latex_form(UnitDef* d)
{
    return (d == NULL ? NULL : d->latex_form);
}
