/**
The Unity package provides a parser for unit strings.

<p><strong>NOTE: The library should currently be regarded as alpha quality:
the implementation and interface may change in response to experience
and comments.</strong></p>

<p>This is the unity parser (version 0.5), which is a Java class library to help parse unit
specification strings such as <code>W mm^-2</code>.  There is also
an associated C library, which uses the same grammars.</p>

<h2>Parsing Units</h2>
You can parse units using a couple of different syntaxes since,
unfortunately, there is no general consensus on which syntax the world
should agree on.  The ones supported (and their names within this
library) are as follows:
<dl>
<dt>fits</dt>
<dd>FITS v3.0, section 4.3, W.D. Pence et al., A&amp;A 524, A42, 2010
        <a href='http://dx.doi.org/10.1051/0004-6361/201015362'>doi:10.1051/0004-6361/201015362</a></dd>
<dt>ogip</dt>
<dd><a href='ftp://legacy.gsfc.nasa.gov/fits_info/fits_formats/docs/general/ogip_93_001/ogip_93_001.ps'
>OGIP memo OGIP/93-001, 1993</a></dd>
<dt>cds</dt>
<dd><a href='http://cdsweb.u-strasbg.fr/doc/catstd-3.2.htx'
>Standards for Astronomical Catalogues, Version 2.0, section 3.2, 2000</a></dd>
</dl>

<p>See also the
<a href='http://www.iau.org/static/publications/stylemanual1989.pdf'
>IAU style manual, section 5.1, 1989</a>, though this is by now rather old.</p>

<p>Each of these has an associated <em>writer</em>, whcih allows you to
write a parsed UnitExpression to a string, in a format which should be
conformant with the particular syntax's standard.
See {@link uk.me.nxg.unity.UnitExpr#toString}.</p>

<p>In addition, there is a <em>latex</em> writer, which produces a
formatted form for the the expression, in a form suitable for
inclusion in a LaTeX document, using the <code>siunitx</code>
package.  To use the resulting output in a LaTeX document, include the
following in the preamble of the file:</p>
<pre>
&#92;usepackage{siunitx}
&#92;DeclareSIQualifier&#92;solar{$&#92;odot$}
</pre>
<p>You may add any <code>siunitx</code> options that seem convenient,
and you may omit the declaration of <code>\solar</code> if the units
in the document do not include the various solar ones.</p>

<p>The parsing is permissive, to the extent that it permits non-recognised and deprecated
units.  The result of the parse may be checked for conformance with
one or other standard using the methods
{@link uk.me.nxg.unity.UnitExpr#allUnitsRecognised},
{@link uk.me.nxg.unity.UnitExpr#allUnitsRecommended} and
{@link uk.me.nxg.unity.UnitExpr#allUsageConstraintsSatisfied}.
Note that SI prefixes are still noticed for unrecognised units: thus <code>furlongs/fortnight</code>
will be parsed as femto-urlongs per femto-ortnight.  The same is not
true of recognised units: a <code>pixel/s</code> is a pixel per
second, and does not involve a pico-ixel.</p>

<h3>Demo</h3>

If you want to experiment with the library, you can do so using the jar file:
<pre>
    % java -jar unity.jar -icds -oogip 'mm2/s'
    mm**2 /s
    % java -jar unity.jar -icds -ofits -v mm/s
    mm s-1
    check: all units recognised?           yes
    check: all units recommended?          yes
    check: all units satisfy constraints?  yes
    % java -jar unity.jar -ifits -ocds -v 'merg/s'
    merg/s
    check: all units recognised?           yes
    check: all units recommended?          no
    check: all units satisfy constraints?  no
    % java -jar unity.jar -icds -ofits -v 'merg/s'
    merg s-1
    check: all units recognised?           no
    check: all units recommended?          no
    check: all units satisfy constraints?  yes
</pre>
In the latter cases, the <code>-v</code> option <em>validates</em> the input string
against various constraints.  The expression <code>mm/s</code> is completely valid
in all the syntaxes.  In the FITS syntax, the <code>erg</code> is a recognised
unit, but it is deprecated; although it is recognised, it is not
permitted to have SI prefixes.  In the CDS syntax, the <code>erg</code> is neither
recognised nor (a fortiori) recommended; since there are no
constraints on it in this syntax, it satisfies all of them (this
latter behaviour is admittedly slightly counterintuitive).


<h2>Grammars supported</h2>

The three supported grammars have a fair amount in common, but the
differences are nonetheless significant enough that they require
separate grammars.  Important differences are in the number of solidi
they allow in the units specifications, and the symbols they use for
products and powers.

<p>The CDS specification permits non-round factors (that is, factors
which aren't a power of ten).  These are not permitted in this CDS
parser, partly because they're arguably quantities rather than units,
but more practically because it significantly complicates the implementation.</p>

<p>Current limitations:</p>
<ul>
<li>No mathematical functions in FITS or OGIP parsers</li>
<li>No log in CDS parser</li>
<li>Currently ignores some of the odder unit restrictions (such as the OGIP requirement that 'Crab' can have a 'milli' prefix, but no other SI prefixes)</li>
</ul>

The library's home page is (at present) at
<a href='http://www.astro.gla.ac.uk/users/norman/ivoa/unity/'><code>http://www.astro.gla.ac.uk/users/norman/ivoa/unity/</code></a>; the source is on
<a href='https://bitbucket.org/nxg/unity'>bitbucket</a>.

<p>In the grammars below, the terminals are as follows:</p>
<ul>
<li>WHITESPACE: one or more whitespace characters</li>
<li>STAR, DOT: a star or a dot, generally used to indicate multiplication</li>
<li>DIVISION: a slash</li>
<li>STARSTAR, CARET: the former is '**'; both are used to indicate exponentiation</li>
<li>OPEN_P, CLOSE_P: open and close parentheses</li>
<li>INTEGER, FLOAT: numbers; the syntax of FLOAT is
<code>[+-]?[1-9][0-9]*\.[0-9]+</code> -- that is, there are no
exponents allowed</li>
<li>STRING: ...a string</li>
</ul>

<h3>The FITS grammar</h3>

<pre>

input:		  product_of_units 
                | factor product_of_units 
// The FITS spec isn't completely clear on the topic of 
// solidi, saying "Parentheses are used for symbol grouping and are
// strongly recommended whenever the order of operations might be
// subject to misinterpretation" and "The IAU style manual forbids
// the use of more than one solidus (/) character in a units
// string. However, since normal mathematical precedence rules apply
// in this context, more than one solidus may be used but is
// discouraged" (p27).  Therefore, it's not clear whether, for
// example, "kg/m s" should be parsed as "kg m-1 s-1", as "kg m-1 s", or
// forbidden on the grounds that 'normal mathematical precedence
// rules' would forbid it (it's probably arguable all ways, but I
// don't think that 'normal mathematical precedence rules' are going
// to resolve it).  Here, we resolve this by declaring that there can
// be only a single expression to the right of the solidus.  That is,
// we do not have "product_of_units DIVISION product_of_units" here.
//
// Note: it is I think a consequence of this that nothing can be
// successully parsed in two different grammars, with different
// meanings.  If the right-hand-side of the division could be a
// product_of_units, then "kg /m s" would parse in both FITS and OGIP,
// but mean "kg m-1 s-1" in FITS and "kg m-1 s" in OGIP.
                | product_of_units DIVISION unit_expression 
                | factor product_of_units DIVISION unit_expression 
// The FITS spec may or may not be intended to permit "10+3 /m",
// but we don't (because ... for heavens' sake!)
                | DIVISION unit_expression 
                ;

unit_expression:  unit                                 
                | OPEN_P product_of_units CLOSE_P 
                ;

product_of_units: unit_expression 
                | product_of_units product unit_expression 
                ;


// We require the following prefix factors to be powers of ten,
// as an extra-syntactic condition
factor:           INTEGER power numeric_power 
// Represent exponents as a pair of integers, eg 10+3
                | INTEGER INTEGER 
                ;


unit:             STRING       		
                | STRING power numeric_power        
                | STRING numeric_power        
                ;

numeric_power:	  INTEGER 
                | OPEN_P INTEGER CLOSE_P 
                | OPEN_P FLOAT CLOSE_P 
                | OPEN_P INTEGER DIVISION INTEGER CLOSE_P 
                ;

power:		  CARET
                | STARSTAR
                ;

product:	WHITESPACE | STAR | DOT;


</pre>

<h3>The OGIP grammar</h3>

<pre>

input:		  product_of_units 
                | factor product_of_units 
                ;

unit_expression:  unit                                 
                | OPEN_P product_of_units CLOSE_P 
                ;


// We conceive of the product_of_units as a sequence of terms 'times
// expression' or 'dividedby expression', multiplying them together
// after, in the latter case, reciprocating them.
product_of_units: unit_expression
                | division unit_expression 
                | product_of_units product unit_expression 
                | product_of_units division unit_expression 
                ;

// We require the following prefix factors to be powers of ten,
// as an extra-syntactic condition
factor:           INTEGER power numeric_power 
                | FLOAT 
                | INTEGER 
                ;

// OGIP recommends no whitespace after the slash
division: DIVISION | WHITESPACE DIVISION | WHITESPACE DIVISION WHITESPACE | DIVISION WHITESPACE;

unit:             STRING       		
                | STRING power numeric_power        
                ;

numeric_power:	  INTEGER 
                | OPEN_P INTEGER CLOSE_P 
                | OPEN_P FLOAT CLOSE_P 
                | OPEN_P INTEGER DIVISION INTEGER CLOSE_P 
                ;

power:		  STARSTAR;

product:        WHITESPACE | STAR | WHITESPACE STAR | WHITESPACE STAR WHITESPACE | STAR WHITESPACE;


</pre>

<h3>The CDS grammar</h3>

This is quite similar to the OGIP grammar, but with more restrictions
<pre>

input:		  product_of_units 
                | factor product_of_units 
                ;

unit_expression:  unit                                 
                | OPEN_P product_of_units CLOSE_P 
                ;


// We conceive of the product_of_units as a sequence of terms 'times
// expression' or 'dividedby expression', multiplying them together
// after, in the latter case, reciprocating them.
product_of_units: unit_expression
                | division unit_expression 
                | product_of_units product unit_expression 
                | product_of_units division unit_expression 
                ;

// We require the following prefix factors to be powers of ten,
// as an extra-syntactic condition
factor:           INTEGER power numeric_power 
// Represent exponents as a pair of integers, eg 10+3
                | INTEGER INTEGER 
                | FLOAT 
                | INTEGER 
                ;

division: DIVISION

unit:             STRING       		
                | STRING numeric_power        
                ;

numeric_power:	  INTEGER 
                ;

power:		  STARSTAR;

product:	DOT;


</pre>

*/
package uk.me.nxg.unity;
