package uk.me.nxg.unity;

import java.util.List;


/**
 * A single unit.  This is represented as a <em>prefix</em>, a
 * <em>base unit</em> and an <em>exponent</em>.  For example, the
 * string "mm s^-2" will result in two OneUnit instances, one of which
 * will have a base unit of "m" and a prefix power of -3, and the
 * other of which will have a base unit of "s" and an exponent of -2.
 *
 * <p>This represents the abstract unit, shorn of the symbols which
 * were parsed to obtain it.  Thus if, in some odd syntax, the symbol
 * 'x' were used to denote metres, then the OneUnit instance which
 * resulted would be the same as the OneUnit which resulted from a
 * more conventional syntax. 
 *
 * <p>Also, there is a potential ambiguity if a symbol is recognised
 * in one syntax, but not in another.  Thus if the string 'erg' were
 * parsed in a syntax which didn't recognise that, then it would be
 * stored as just that, an unrecognised symbol, not associated with
 * the {@link UnitDefinition} for the erg.
 *
 * <p>The class's instances are immutable.
 */
public class OneUnit {
    /**
     * Represents the power-of-ten modifier in front of the base
     * unit.  For example, in the string "mm", the prefix power would be "-3".
     */
    private final int prefixPower;
    /**
     * The base unit as a UnitDefinition.
     * The 'base unit' is the unit without the prefix.
     * Thus in the expression "MW", it is 'W', Watt, that is the base unit.
     * This doesn't refer to the fundamental SI base units.
     * Precisely one of baseUnitString and baseUnitDefinition is non-null.
     */
    private final UnitDefinition baseUnitDefinition;
    /**
     * The base unit as a string.  If this is non-null, it means that
     * the unit string wasn't recognised as a known-unit.
     */
    private final String baseUnitString;
    /**
     * The power to which this base unit is raised.
     */
    private final double exponent;

    /**
     * We store the resolver which was used to look up the symbols
     * which identified this unit.  This makes it possible to pass
     * this along to the UnitExpr class.
     */
    private final UnitDefinitionMap.Resolver resolver;
    

    /** 
     * A regexp to match a unit with an SI prefix.
     * The order in this regexp is important -- the "da" must match
     * before "d" can, so that "dam" is parsed as "da"+"m", and not
     * "d"+"am".  Also, this regexp requires there to be a unit: "m"
     * is not parsed as a prefix plus an empty unit.
     */
    private static final java.util.regex.Pattern unitWithPrefix
            = java.util.regex.Pattern.compile("(da|[yzafpnumcdhkMGTPEZY])(.+)");

    /**
     * Two units are equal if they have the same power, units and
     * exponent.  They don't have to have been obtained from the same
     * syntax, so that in a syntax which (for example) permitted both
     * "yr" and "a" for years, two OneUnit instances, obtained from
     * parsing the two alternatives, would be regarded as equal.
     */
    @Override public boolean equals(Object o) {
        // Implementation note: the practical effect of the above
        // notes is that the 'resolver' field should be ignored for
        // the purposes of the equals and hashCode methods.
        if (o == this) {
            return true;        // trivially
        } else if (o instanceof OneUnit) {
            OneUnit oo = (OneUnit)o;
            return oo.prefixPower == prefixPower
                    && oo.baseUnitDefinition == baseUnitDefinition
                    && oo.baseUnitString == baseUnitString
                    && oo.exponent == exponent;
        } else {
            return false;
        }
    }

    @Override public int hashCode() {
        // We've overridden equals(), so we have to override hashCode,
        // too (as noted in eg Bloch, Effective Java, prescription 8,
        // and using the simple approach recommended there).
        int result = 17;
        result = 37*result + prefixPower;
        if (baseUnitDefinition == null) {
            assert baseUnitString != null;
            result = 37*result + baseUnitString.hashCode();
        } else {
            assert baseUnitString == null;
            result = 37*result + baseUnitDefinition.hashCode();
        }
        long de = Double.doubleToLongBits(exponent);
        result = 37*result + (int)(de ^ (de >>> 32));
        return result;
    }

    /**
     * Private internal constructor
     */
    private OneUnit(int prefix, String us, UnitDefinition ud, double e, UnitDefinitionMap.Resolver resolver) {
        prefixPower = prefix;
        if (us == null) {
            baseUnitString = null;
            baseUnitDefinition = ud;
        } else {
            baseUnitString = us;
            baseUnitDefinition = null;
        }
        exponent = e;
        this.resolver = resolver;

        // class invariant
        assert (baseUnitString == null && baseUnitDefinition != null)
                || (baseUnitString != null && baseUnitDefinition == null);
        assert resolver != null;
    }

    /**
     * Constructor of base unit -- no multiple.
     */
    private OneUnit(UnitDefinitionMap.Resolver resolver, String s, int e) {
        this(resolver, s, (double)e);
    }

    /**
     * Constructor of base unit
     * @param unitString a string representation of a single unit with
     * a possible SI prefix
     * @param e the power to which this unit is raised
     */
    private OneUnit(UnitDefinitionMap.Resolver resolver, String unitString, double e) {
        // It would be possible/appropriate to do strict checking of
        // units in here -- that is, throwing exceptions whenever we
        // encounter any non-recommended unit (ie one which lookupUnit
        // returns null for).  This would create the additional
        // possibility of being a convenient/reasonable place to check
        // whether the abbreviation that we look up is an allowed one
        // for the syntax we're parsing, as long as we were
        // additionally provided with some information within this
        // constructor, about which syntax that is (which is currently
        // not available in here).
        UnitDefinition def;
        if ((def = resolver.lookupSymbol(unitString)) != null) {
            prefixPower = 0;
            baseUnitString = null;
            baseUnitDefinition = def;
        } else {
            java.util.regex.Matcher m = unitWithPrefix.matcher(unitString);
            if (m.matches()) {
                prefixPower = prefixLetterToPower(m.group(1).equals("da")
                                                  ? UnitVal.CHAR_DECA
                                                  : m.group(1).charAt(0));
                def = resolver.lookupSymbol(m.group(2));
                if (def == null) {
                    baseUnitString = m.group(2);
                    baseUnitDefinition = null;
                } else {
                    baseUnitString = null;
                    baseUnitDefinition = def;
                }
            } else {
                prefixPower = 0;
                baseUnitString = unitString;
                baseUnitDefinition = null;
            }
        }
        exponent = e;
        this.resolver = resolver;

        // class invariant
        assert (baseUnitString == null && baseUnitDefinition != null) || (baseUnitString != null && baseUnitDefinition == null);
        assert resolver != null;
    }

    /**
     * Produce the reciprocal of the unit, which is a unit which has
     * the negative of the exponent of this one.
     * @return a new instance
     */
    OneUnit reciprocate() {
        return new OneUnit(prefixPower, baseUnitString, baseUnitDefinition, -exponent, resolver);
    }

    /**
     * Returns the prefix of the unit, as a base-ten log.  Thus a
     * prefix of "m", for "milli", would produce a prefix of -3.
     */
    public int getPrefix() {
        return prefixPower;
    }

    /**
     * Returns the power to which this unit is raised.
     * A unit of "mm^3" would return 3.0, and "m^(-1/2)" would return -0.5.
     */
    public double getPower() {
        return exponent;
    }    
    
    /**
     * Convert SI prefixes to the decimal power they correspond to
     */
    private static int prefixLetterToPower(char prefixLetter) {
        int prefixPower;
        switch (prefixLetter) {
          case 'y': prefixPower = -24; break;
          case 'z': prefixPower = -21; break;
          case 'a': prefixPower = -18; break;
          case 'f': prefixPower = -15; break;
          case 'p': prefixPower = -12; break;
          case 'n': prefixPower =  -9; break;
          case 'u': prefixPower =  -6; break;
          case 'm': prefixPower =  -3; break;
          case 'c': prefixPower =  -2; break;
          case 'd': prefixPower =  -1; break;
          case UnitVal.CHAR_DECA:
            prefixPower = 1;
            break;
          case 'h': prefixPower =   2; break;
          case 'k': prefixPower =   3; break;
          case 'M': prefixPower =   6; break;
          case 'G': prefixPower =   9; break;
          case 'T': prefixPower =  12; break;
          case 'P': prefixPower =  15; break;
          case 'E': prefixPower =  18; break;
          case 'Z': prefixPower =  21; break;
          case 'Y': prefixPower =  24; break;
          default:
            // The lexer in lex.jflex should make this impossible
            throw new AssertionError("Impossible prefix: '" + prefixLetter + "'");
        }
        return prefixPower;
    }

    /**
     * Convert a prefix power to a String.  These are returned as
     * Strings, so that we can later consider 'da' or 'ki' (etc) as prefixes.
     */
    private static String prefixPowerToString(int prefixPower) {
        String rval;
        switch (prefixPower) {
          case -24: rval = "y"; break;
          case -21: rval = "z"; break;
          case -18: rval = "a"; break;
          case -15: rval = "f"; break;
          case -12: rval = "p"; break;
          case  -9: rval = "n"; break;
          case  -6: rval = "u"; break;
          case  -3: rval = "m"; break;
          case  -2: rval = "c"; break;
          case  -1: rval = "d"; break;
          case   0: rval = "";  break;
          case   1: rval = "da";break;
          case   2: rval = "h"; break;
          case   3: rval = "k"; break;
          case   6: rval = "M"; break;
          case   9: rval = "G"; break;
          case  12: rval = "T"; break;
          case  15: rval = "P"; break;
          case  18: rval = "E"; break;
          case  21: rval = "Z"; break;
          case  24: rval = "Y"; break;
          default:
            // This should be impossible, since it's only
            // prefixLetterToPower which sets these values
            throw new AssertionError("Impossible prefix power: " + prefixPower);
        }
        return rval;
    }

    /**
     * Convert a prefix power to a String, in a form useful within
     * siunitx.  The only real difference from {@link
     * prefixPowerToString} is that <code>\micro</code> is formatted differently.
     */
    private static String prefixPowerToStringLaTeX(int prefixPower) {
        String rval;
        switch (prefixPower) {
          case -24: rval = "\\yocto"; break;
          case -21: rval = "\\zepto"; break;
          case -18: rval = "\\atto";  break;
          case -15: rval = "\\femto"; break;
          case -12: rval = "\\pico";  break;
          case  -9: rval = "\\nano";  break;
          case  -6: rval = "\\micro"; break;
          case  -3: rval = "\\milli"; break;
          case  -2: rval = "\\centi"; break;
          case  -1: rval = "\\deci";  break;
          case   0: rval = "";        break;
          case   1: rval = "\\deca";  break;
          case   2: rval = "\\hecto"; break;
          case   3: rval = "\\kilo";  break;
          case   6: rval = "\\mega";  break;
          case   9: rval = "\\giga";  break;
          case  12: rval = "\\tera";  break;
          case  15: rval = "\\peta";  break;
          case  18: rval = "\\exa";   break;
          case  21: rval = "\\zetta"; break;
          case  24: rval = "\\yotta"; break;
          default:
            // This should be impossible, since it's only
            // prefixLetterToPower which sets these values
            throw new AssertionError("Impossible prefix power: " + prefixPower);
        }
        return rval;
    }

    static List<OneUnit> divide(List<OneUnit> num, Iterable<OneUnit> den) {
        List<OneUnit> res = new java.util.ArrayList<OneUnit>(num);
        for (OneUnit u : den) {
            res.add(new OneUnit(u.prefixPower, u.baseUnitString, u.baseUnitDefinition, -u.exponent, u.resolver));
        }
        return res;
    }

    static List<OneUnit> reciprocate(Iterable<OneUnit> ul) {
        List<OneUnit> res = new java.util.ArrayList<OneUnit>();
        for (OneUnit u : ul) {
            res.add(u.reciprocate());
        }
        return res;
    }

    /**
     * Format this unit in some sort of canonical form.  The form is
     * unspecified.  This should
     * not generally be used for formatted output -- for that, it will
     * generally be more appropriate to use {@link UnitExpr#toString}.
     */
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append(unitString());
        if (exponent != 1) {
            sb.append('^').append(exponent);
        }
        return sb.toString();
    }

    /**
     * Obtains the string representation of the unit, including
     * prefix, in the given syntax.  That is, return 'mm' not 'm' or 'metre'.
     * @param syntaxName one of the syntaxes of UnitParser
     * @return a non-null string representation of the unit
     * @throws UnitParserException if the syntax is unrecognised
     */
    public String unitString(String syntaxName)
            throws UnitParserException {
        String ret;
        if (UnitParser.LATEX.equals(syntaxName)) {
            ret = unitStringLaTeX();
        } else {
            ret = unitStringText(syntaxName);
        }
        assert ret != null;
        return ret;
    }

    /**
     * Returns the known base unit.  If the unit wasn't recognised as a
     * known unit in the syntax in which the string was parsed, then
     * this returns null, though {@link #getBaseUnitString} will not.
     *
     * <p>Note that the 'base unit' is simply the unit without the prefix, and
     * doesn't refer to the fundamental SI base units.  Thus in the
     * expression "MW", it is 'W', Watt, that is the base unit.
     */
    public UnitDefinition getBaseUnitDefinition() {
        return baseUnitDefinition;
    }

    /**
     * Returns the name of this unit.  If the unit is a known one,
     * then this will return a name for the unit such as 'Metre', or
     * 'Julian year'; if it is not, then this can do no more than
     * return the unit symbol.
     */
    public String getBaseUnitName() {
        assert baseUnitString != null || baseUnitDefinition != null;
        return (baseUnitDefinition != null ? baseUnitDefinition.name() : baseUnitString);
    }

    /** 
     * Returns the base unit string, which will only be non-null if
     * this unit was an unrecognised one.  This is non-private
     * because UnitExpr needs to know this when it's searching for
     * units in the expression; it's package-only because this is not
     * how clients should format units -- they should use the UnitExpr
     * formatting facilities instead.
     */
    String getBaseUnitString() {
        return baseUnitString;
    }

    /**
     * Return the dimensions of the unit, if it is a recognised one.
     * If this <em>isn't</em> a recognised unit, return null.
     * @return the dimensions of the unit, or null if these aren't avaiable
     */
    public Dimensions getDimensions() {
        if (baseUnitDefinition == null) {
            return null;
        } else {
            return baseUnitDefinition.dimensions();
        }
    }
    
    /**
     * Retrieve the string form of the unit in one of the text syntaxes.
     * @param syntaxName one of the syntaxes of {@link UnitParser}
     * @return the text unit as, for example, "mm" for millimetres
     * @throws UnitParserException if the syntax is unrecognised
     */
    private String unitStringText(String syntaxName)
            throws UnitParserException {
        StringBuilder sb = new StringBuilder();
        if (prefixPower != 0) {
            sb.append(prefixPowerToString(prefixPower));
        }
        if (baseUnitString != null) {
            sb.append(baseUnitString);
        } else {
            UnitDefinitionMap.Resolver r;
            if (syntaxName == null) {
                // special call, from unitString(void)
                r = resolver;
            } else {
                r = UnitDefinitionMap.getInstance().getResolver(syntaxName);
                if (r == null) {
                    throw new UnitParserException("Unknown syntax: " + syntaxName);
                }
            }
            assert baseUnitDefinition != null;
            UnitRepresentation rep = r.lookupUnit(baseUnitDefinition);
            if (rep == null) {
                // this UnitDefinition is not recognised in syntax 'syntaxName'
                rep = resolver.lookupUnit(baseUnitDefinition);
                // this resolver is where we originally got the
                // baseUnitDefinition from, so this must succeed
            }
            assert rep != null;
            sb.append(rep.getAbbreviation());
        }
        return sb.toString();
    }

    /**
     * Retrieve the string form of the unit in a form suitable for use
     * within the siunitx LaTeX package.  The LaTeX serialisation
     * isn't quite a syntax like the text ones, though it appears
     * similar to one from the outside.
     * @return the text unit as, for example, "mm" for millimetres
     * @throws UnitParserException if the syntax is unrecognised
     */
    private String unitStringLaTeX()
            throws UnitParserException {
        StringBuilder sb = new StringBuilder();
        if (prefixPower != 0) {
            sb.append(prefixPowerToStringLaTeX(prefixPower)).append(' ');
        }
        if (baseUnitString != null) {
            sb.append(baseUnitString);
        } else {
            assert baseUnitDefinition != null; // from class invariant

            if (baseUnitDefinition.latexForm() != null) {
                sb.append(baseUnitDefinition.latexForm());
            } else {
                UnitRepresentation rep = resolver.lookupUnit(baseUnitDefinition);
                // this resolver is where we originally got the
                // baseUnitDefinition from, so this must succeed
                assert rep != null;

                sb.append(rep.getAbbreviation());
            }
        }
        return sb.toString();
    }
    
    /**
     * Obtains the string form of the unit, including prefix, with a
     * default syntax.  That is, return 'mm' not 'm'.
     *
     * <p>The default syntax is (currently) the syntax with which this
     * unit was originally read.
     */
    public String unitString() {
        try {
            return unitString(null);
        } catch (UnitParserException e) {
            throw new AssertionError("Odd: OneUnit.unitString() should not have failed");
        }
    }
    
    /**
     * Obtains the power the unit is raised to.  For example, for the
     * unit 'mm^2', return 2.
     */
    public double getExponent() {
        return exponent;
    }

    /**
     * Indicates whether the base unit is one of those recognised
     * within the specification of the given syntax.  In this context,
     * "recognised" means "mentioned in the specification", so deprecated units
     * count as recognised ones.
     *
     * <p>Note that this checks that the <em>unit</em> is a
     * recommended one: we don't (currently) check whether the
     * abbreviation that got us here is a recommended one (for
     * example, "pixel" is a valid FITS/CDS name for pixels, and "pix"
     * is a FITS and OGIP one).
     * @param syntax one of the syntaxes of {@link UnitParser}
     * @see #isRecommendedUnit
     */
    public boolean isRecognisedUnit(String syntax) {
        // We could check the validity of the abbreviation whilst we're
        // parsing the units, but that would frustrate our ability to
        // subsequently write out recommended units in the appropriate
        // form.
        return baseUnitDefinition != null
                && baseUnitDefinition.getRepresentation(syntax) != null;
    }

    /**
     * Indicates whether the base unit is one of those recommended
     * within the specification of the given syntax.  In this context,
     * "recommended" means "mentioned in the specification"
     * <em>and</em> not deprecated.
     *
     * <p>Note that this checks that the <em>unit</em> is a
     * recommended one: we don't (currently) check whether the
     * abbreviation that got us here is a recommended one (for
     * example, "pixel" is a valid FITS/CDS name for pixels, and "pix"
     * is a FITS and OGIP one).
     * @param syntax one of the syntaxes of {@link UnitParser}
     * @see #isRecognisedUnit
     */
    public boolean isRecommendedUnit(String syntax) {
        // We could check the validity of the abbreviation, when we're
        // parsing the units, but that would frustrate our ability to
        // subsequently write out recommended units in the appropriate
        // form.
        if (baseUnitDefinition == null) {
            return false;
        } else {
            UnitRepresentation r = baseUnitDefinition.getRepresentation(syntax);
            return r != null && !r.isDeprecated();
        }
    }

    /**
     * Indicates whether the unit is being used in a way which
     * satisfies any usage constraints.  Principally, this tests
     * whether a unit which may not be used with SI prefixes was
     * provided with a prefix, but there may be other constraints
     * present.
     *
     * <p>An unrecognised unit has no constraints, and so will always
     * satisfy them; this extends to units which are unrecognised in a
     * particular syntax.
     * @param syntax one of the syntaxes of {@link UnitParser}
     */
    public boolean satisfiesUsageConstraints(String syntax) {
        // Other constraints possible: for some syntaxes, the 'crab'
        // unit may be used with only 'm' prefix.
        boolean validity;
        if (baseUnitDefinition == null) {
            // there are no constraints, so this necessarily satisfies all of them
            validity = true;
        } else {
            UnitRepresentation r = baseUnitDefinition.getRepresentation(syntax);
            if (r == null) {
                // as above
                validity = true;
            } else {
                validity = prefixPower == 0 || r.mayHaveSIPrefixes();
            }
        }
        return validity;
    }

    /**
     * Provides the resolver which was used to look up symbols when
     * constructing this unit.  This is deliberately package-only --
     * there should be no need for clients outside this package to
     * know this.
     */
    UnitDefinitionMap.Resolver getResolver() {
        return resolver;
    }

    /**
     * Write out the unit in a testable format.
     * This is for debugging and testing.
     */
    public String toDebugString() {
        StringBuilder sb = new StringBuilder();
        sb.append(prefixPower).append('/');
        if (baseUnitString == null) {
            sb.append(baseUnitDefinition.getRepresentation(UnitParser.FITS).getAbbreviation());
        } else {
            sb.append(baseUnitString);
        }
        sb.append('/');
        if (exponent == Math.floor(exponent)) { // no fractional part
            sb.append(Math.round(exponent)); // print an integer
        } else {
            sb.append(exponent);
        }
        return sb.toString();
    }

    // Helper methods

    /**
     * A factory for OneUnit instances.  This is provided so that the
     * parser can construct these instances without having to be
     * given, or care about, the resolver object.
     */
    static class Maker {
        final UnitDefinitionMap.Resolver resolver;
        Maker(UnitDefinitionMap.Resolver resolver) {
            if (resolver == null) {
                throw new IllegalArgumentException("Can't make OneUnit.Maker with null resolver");
            }
            this.resolver = resolver;
        }
        /** Construct a OneUnit from a string unit and an integer
         * exponent.  For example, cubic millimetres would have
         * arguments "mm" and 3.
         */
        public OneUnit make(String s, int e) {
            return new OneUnit(resolver, s, e);
        }
        /** Construct a OneUnit from a string unit and a real
         * exponent.  For example, cubic millimetres would have
         * arguments "mm" and 3.0.
         */
        public OneUnit make(String s, double e) {
            return new OneUnit(resolver, s, e);
        }
    }
}
