package uk.me.nxg.unity;

/**
 * A factory which associates string parser types with actual parsers.
 */
class SyntaxFactory {

    // closures are bloody hard in Java
    private static final java.util.Map<String,SyntaxMaker> parsers;
    static {
        java.util.HashMap<String,SyntaxMaker> m = new java.util.HashMap<String,SyntaxMaker>();
        m.put(UnitParser.FITS, new SyntaxMaker() {
                public Parser makeParser() { return new Parser_fits(); }
                public UnitWriter makeWriter(final UnitExpr e) { return new UnitWriterFits(e); }
            });
        m.put(UnitParser.OGIP, new SyntaxMaker() {
                public Parser makeParser() { return new Parser_ogip(); }
                public UnitWriter makeWriter(final UnitExpr e) { return new UnitWriterOgip(e); }
            });
        m.put(UnitParser.CDS, new SyntaxMaker() {
                public Parser makeParser() { return new Parser_cds(); }
                public UnitWriter makeWriter(final UnitExpr e) { return new UnitWriterCds(e); }
            });
        m.put(UnitParser.LATEX, new SyntaxMaker() {
                public Parser makeParser()
                        throws UnitParserException {
                    throw new UnitParserException("Can't create a LaTeX parser");
                }
                public UnitWriter makeWriter(final UnitExpr e) { return new UnitWriterLaTeX(e); }
            });
        m.put("debug", new SyntaxMaker() {
                public Parser makeParser()
                        throws UnitParserException {
                    throw new UnitParserException("Can't create a debug parser");
                }
                public UnitWriter makeWriter(final UnitExpr e) { return new UnitWriterDebugging(e); }
            });
        parsers = java.util.Collections.unmodifiableMap(m);
    }

    private SyntaxFactory() {
        // does nothing -- prevent callers constructing instance
    }

    /**
     * Create a new parser, to parse a single input.
     * @param syntax the type of parser required -- one of the parser strings returned by {@link getParserTypes}
     * @param input the string to be parsed
     * @return a Parser which can be called to parse the given input,
     * or null if there is no parser with the specified name
     */
    static Parser createParser(String syntax, java.io.Reader input)
            throws UnitParserException {
        Parser parser;

        SyntaxMaker pm = parsers.get(syntax);
        if (pm == null) {
            parser = null;
        } else {
            UnitDefinitionMap.Resolver resolver = UnitDefinitionMap.getInstance().getResolver(syntax);
            // if pm isn't null, it should be impossible for resolver to be null
            assert resolver != null;

            parser = pm.makeParser();
            parser.setUnitMaker(new OneUnit.Maker(resolver));
            parser.setReader(input);
        }
        return parser;
    }

    /**
     * Creates a writer for a single unit-expression
     * @param type the syntax of the string to be written -- one of the parser strings returned by {@link getParserTypes}
     * @param ue the single unit-expressin which is to be written out
     * @return a writer
     * @throws UnitParserException if the writer cannot be created (typically because the syntax is unknown)
     */
    static UnitWriter createWriter(String type, UnitExpr ue)
            throws UnitParserException {
        SyntaxMaker pm = parsers.get(type);
        if (pm == null) {
            throw new UnitParserException("Unknown syntax: " + type);
        } 
        return pm.makeWriter(ue);
    }

    /**
     * Indicate the set of available parsers.
     * The returned syntaxes are all valid as arguments to {@link createParser}
     * @return a list of parser names
     */
    static java.util.Set<String> getParserTypes() {
        return parsers.keySet();
    }

    /**
     * Indicates whether the given parser is a known one.
     * @param parser the string name of a parser (eg "fits")
     * @return true if the parser is one of the types returned by
     * {@link getParserTypes}
     */
    public static boolean isKnownParser(String parser) {
        return parsers.get(parser) != null;
    }
}
